import os
from sys import argv
import ujson as json
import pandas as pd
from io import StringIO
from pathlib import Path
from functools import partial
from p_tqdm  import p_uimap
from argparse import ArgumentParser, RawDescriptionHelpFormatter

from eis1600.repositories.repo import get_ready_and_double_checked_files, TEXT_REPO, JSON_REPO, RECONSTRUCTED_REPO, \
                                       RECONSTRUCTED_INFIX
from eis1600.helper.CheckFileEndingActions import CheckFileEndingEIS1600JsonAction
from eis1600.helper.fix_dataframe import add_missing_columns
from eis1600.processing.postprocessing import write_updated_miu_to_file
from eis1600.yml.YAMLHandler import YAMLHandler


TAGS_LIST_FOR_RECONSTRUCTION = ('MSS', 'PAGES', 'DATE_TAGS', 'MONTH_TAGS', 'ONOM_TAGS', 'ONOMASTIC_TAGS', 'NER_TAGS')


def reconstruct_file(
        fpath: str,
        tags_list: tuple[str],
        force: bool = False,
        add_annotations_yml: bool = False,
        ):

    if fpath.endswith(".json"):
        out_fpath = fpath.replace(".json", f"{RECONSTRUCTED_INFIX}.EIS1600")
    else:
        fpath = fpath.replace(TEXT_REPO, JSON_REPO)
        out_fpath = fpath.replace(".EIS1600", f"{RECONSTRUCTED_INFIX}.EIS1600")
        fpath = fpath.replace('.EIS1600', '.json')

    out_fpath = out_fpath.replace(JSON_REPO, RECONSTRUCTED_REPO)

    if not Path(fpath).is_file():
        print(f"Warning! file {fpath} not found")
        return

    # do not process file if it's already generated and it should not be overwritten
    if Path(out_fpath).is_file() and not force:
        return

    dir_path, _ = os.path.split(out_fpath)
    Path(dir_path).mkdir(parents=True, exist_ok=True)

    with open(fpath, "r", encoding="utf-8") as fp, \
         open(out_fpath, "w", encoding="utf-8") as outfp:
        data = json.load(fp)

        yml = data[0]["yml"]
        yml_handler = YAMLHandler(yml, ignore_annotations=not add_annotations_yml)
        df = pd.concat([pd.read_json(StringIO(miu["df"])) for miu in data], ignore_index=True)
        df = add_missing_columns(df)
        df["TAGS_LISTS"] = None
        if 'ONOM_TAGS' in df:
            df['ONOM_TAGS'] = df['ONOM_TAGS'].map(lambda s: '' if s == '_' else s)
        write_updated_miu_to_file(
            outfp,
            yml_handler,
            df,
            target_columns=tags_list,
            forced_re_annotation=True,
            add_annotations_to_yml=add_annotations_yml
        )


def main():
    arg_parser = ArgumentParser(
            prog=argv[0], formatter_class=RawDescriptionHelpFormatter,
            description="Script to reconstruct an EIS1600 text file form the json output."
    )
    arg_parser.add_argument(
            'infile', type=str, nargs='?',
            help='json file to process',
            action=CheckFileEndingEIS1600JsonAction
    )
    arg_parser.add_argument(
            '--tags', nargs='+',
            choices=TAGS_LIST_FOR_RECONSTRUCTION,
            default=TAGS_LIST_FOR_RECONSTRUCTION,
            help='list of tags to include in the reconstructed text. All by default'
    )
    arg_parser.add_argument(
            '--force', action='store_true',
            help='create file even though it is already created'
    )
    arg_parser.add_argument(
            '--add_annotations_to_yml', action='store_true',
            help='add all annotations to yml header'
    )
    args = arg_parser.parse_args()

    tags_list = tuple(args.tags)

    if args.infile:
        reconstruct_file(args.infile, tags_list=tags_list, force=True)

    else:
        files_ready, files_double_checked = get_ready_and_double_checked_files()
        files = files_ready + files_double_checked

        list(p_uimap(partial(reconstruct_file,
                             tags_list=tags_list,
                             force=args.force,
                             add_annotations_yml=args.add_annotations_to_yml
                             ),
                     files,
                     num_cpus=0.7)
             )

        print(f"Reconstructed files have been saved in {RECONSTRUCTED_REPO} directory.")



import os.path
import jsonpickle
import ujson as json
import pandas as pd
from pathlib import Path
from sys import argv
from tqdm import tqdm
from argparse import ArgumentParser, RawDescriptionHelpFormatter

from eis1600.repositories.repo import get_ready_and_double_checked_files, TEXT_REPO, JSON_REPO, TSV_REPO, COLUMNS, \
    SEP, SEP2
from eis1600.helper.CheckFileEndingActions import CheckFileEndingEIS1600JsonAction


ALL_LABELS = ("SECTIONS", "TOKENS", "TAGS_LISTS", "NER_LABELS", "LEMMAS", "POS_TAGS", "ROOTS", "TOPONYM_LABELS",
              "NER_TAGS", "DATE_TAGS", "MONTH_TAGS", "ONOM_TAGS", "ONOMASTIC_TAGS")


def dump_file(fpath: str, label_list: tuple[str] = ALL_LABELS):

    if not fpath.endswith(".json"):
        fpath = fpath.replace(TEXT_REPO, JSON_REPO)
        fpath = fpath.replace('.EIS1600', '.json')

    structural_data, content_data = [], []
    with open(fpath, "r", encoding="utf-8") as fp:
        data = json.load(fp)

    for miu in data:
        header = miu["yml"]
        uid = header["UID"]

        # get structural data
        for entity in header.keys():
            if entity == "UID":
                continue
            value = header[entity]
            if type(value) in (str, int, bool):
                structural_data.append((uid, entity, value))
            elif type(value) == list:
                for sub_value in value:
                    if type(sub_value) == dict:
                        parsed_sub_value = SEP.join(f"{k}{SEP2}{v}" for k, v in sub_value.items())
                        structural_data.append((uid, entity, parsed_sub_value))
                    elif type(sub_value) == list:
                        parsed_sub_value = SEP.join(sub_value)
                        structural_data.append((uid, entity, parsed_sub_value))
                    else:
                        structural_data.append((uid, entity, sub_value))
            elif type(value) == dict:
                for sub_entity, sub_value in value.items():
                    if type(sub_value) == list:
                        for val in sub_value:
                            structural_data.append((uid, entity, f"{sub_entity}{SEP}{val}"))
                    elif type(sub_value) == dict:
                        # e.g. onomastics elements
                        for sub_sub_ent, sub_sub_val in sub_value.items():
                            structural_data.append((uid, entity, f"{sub_entity}{SEP}{sub_sub_ent}{SEP2}"))
                    else:
                        structural_data.append((uid, entity, f"{sub_entity}{SEP}{sub_value}"))
            else:
                raise ValueError(f'Fatal error dumping data of "{fpath}".')

        # get content data
        miu_df = jsonpickle.decode(miu["df"])
        for entity in label_list:
            if entity not in miu_df:
                continue
            for j, (_, value) in enumerate(miu_df[entity].items(), 1):
                if type(value) == list:
                    value = SEP2.join(value)
                if value:
                    content_data.append((uid, entity, f"{j}{SEP}{value}"))

    fbase, _ = os.path.splitext(fpath)

    fbase = fbase.replace(JSON_REPO, TSV_REPO)
    dir_path, _ = os.path.split(fbase)
    Path(dir_path).mkdir(parents=True, exist_ok=True)

    content_df = pd.DataFrame(content_data, columns=COLUMNS)
    content_df.to_csv(f"{fbase}_df.tsv", sep="\t", index=False)

    struct_df = pd.DataFrame(structural_data, columns=COLUMNS)
    struct_df.to_csv(f"{fbase}_yml.tsv", sep="\t", index=False)


def main():
    arg_parser = ArgumentParser(
            prog=argv[0], formatter_class=RawDescriptionHelpFormatter,
            description="Script to dump all eis1600 corpus into a tsv with the structure and "
                        "another tsv file with the enriched textual information."
    )
    arg_parser.add_argument(
            'infile', type=str, nargs='?',
            help='json file to process',
            action=CheckFileEndingEIS1600JsonAction
    )
    arg_parser.add_argument(
        "--label_list",
        nargs="*",
        default=ALL_LABELS,
        help="entities from content data to add to output. The default is all entities: "
             "SECTIONS, TOKENS, TAGS_LISTS, NER_LABELS, LEMMAS, POS_TAGS, ROOTS, TOPONYM_LABELS, "
             "NER_TAGS, DATE_TAGS, MONTH_TAGS, ONOM_TAGS, ONOMASTIC_TAGS"
    )
    args = arg_parser.parse_args()

    if args.infile:
        dump_file(args.infile, args.label_list)

    else:
        files_ready, files_double_checked = get_ready_and_double_checked_files()
        files = files_ready + files_double_checked

        for fpath in tqdm(files):

            dump_file(fpath, args.label_list)

        print(f"Processed {len(files)} files")
        print(f"For each json file in {JSON_REPO} directory, "
              f"a tsv with the yml data and a tsv with the df data have been generated.")





from enum import Enum
from collections import namedtuple


BIO_MAN = "$BIO_MAN$"
BIO_WOM = "$BIO_WOM$"
BIO_REP = "$BIO_REP$"
BIO_LST = "$BIO_LST$"

BIO_TAGS = (BIO_MAN, BIO_WOM, BIO_REP, BIO_LST)
EVENT_TAGS = ("@", "|", "||", "|||", "||||", "|||||", "||||||", "|||||||", "||||||||", "|||||||||", "||||||||||")
OTHER_TAGS = ("$DIC_BIB$", "$DIC_LEX$", "$DIC_NIS$", "$DIC_TOP$")
EXTERNAL = ("|APPENDIX|", "|PARATEXT|", "|EDITOR|")


class CategoryType(Enum):
    EVENT = "event"
    BIOGRAPHY = "biography"
    OTHER = "other"
    EXTERNAl = "external"


class Category:

    type: CategoryType = None
    value: str = None

    def __init__(self, value: str):
        if value in BIO_TAGS:
            category = CategoryType.BIOGRAPHY
        elif value in EVENT_TAGS:
            category = CategoryType.EVENT
        elif value in OTHER_TAGS:
            category = CategoryType.OTHER
        elif value in EXTERNAL:
            category = CategoryType.EXTERNAl
        else:
            raise ValueError(f"category {value} not recognised")
        self.value = value
        self.type = category

    def is_bio(self) -> bool:
        return self.type == CategoryType.BIOGRAPHY


def convert_to_longer_bio_tag(value: str) -> str:
    if any(tag in value for tag in BIO_TAGS+OTHER_TAGS):
        return value
    if "$$$$" in value:
        value = value.replace("$$$$", BIO_LST)
    elif "$$$" in value:
        value = value.replace("$$$", BIO_REP)
    elif "$$" in value:
        value = value.replace("$$", BIO_WOM)
    elif "$" in value:
        value = value.replace("$", BIO_MAN)
    return value



from typing import Any

import pandas as pd

def fix_bonom_position(df: pd.DataFrame) -> pd.DataFrame:
    """
    BONOM tag is sometimes added to an empty section below. It should be moved to the next section.
        Example input:
            SECTION      TOKEN      TAG
            section_a    tok1       None
            None         None       MYTAG
            section_b    tok2       None
            None         tok3       None
            None         tok4       None
        Example output:
            SECTION      TOKEN      TAG
            section_a    tok1       None
            None         None       None
            section_b    tok2       MYTAG
            None         tok3       None
            None         tok4       None
    """
    # find the indices of rows where ONOM_TAG is BONOM
    mytag_indices = df[df['ONOM_TAGS'] == 'BONOM'].index

    # create masks for previous and next rows
    prev_mask = mytag_indices - 1 >= 0
    next_mask = mytag_indices + 1 < len(df)

    prev_section_not_none = df.loc[mytag_indices - 1, 'SECTIONS'].notna().values
    next_section_not_none = df.loc[mytag_indices + 1, 'SECTIONS'].notna().values

    misplaced_bonom = prev_mask & next_mask & prev_section_not_none & next_section_not_none
    df.loc[mytag_indices[misplaced_bonom], 'ONOM_TAGS'] = None
    df.loc[mytag_indices[misplaced_bonom] + 1, 'ONOM_TAGS'] = 'BONOM'

    return df


def _extract_value(tags_list: list, value: str) -> str | None:
    for tag in tags_list:
        if value in tag:
            return tag
    return None


def add_missing_columns(df: pd.DataFrame) -> pd.DataFrame:
    """
    Add a column with the pages tags so that they are not lost when the text is reconstructed.
    Potentially other tags can be easily added just by including the pair (COL, VAl) below.
    """
    for col, val in (("PAGES", "Page"),):
        df[col] = df['TAGS_LISTS'].apply(
            lambda x: _extract_value(x, val) if x is not None and isinstance(x, list) else None
        )
    return df
